package de.klausj.video.impl.gstreamer;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.PixelInterleavedSampleModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.net.URI;
import java.nio.ByteBuffer;
import java.util.Hashtable;
import java.util.Vector;

import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;

import ips.media.FrameStepControl;
import ips.media.MediaDecodingPlayer;
import ips.media.NativeMediaResult;
import ips.media.NativeMediaSystemException;
import ips.media.SeekControl;
import ips.media.control.MediaControlListener;
import ips.media.event.MediaCloseEvent;
import ips.media.event.MediaControlEventTransferAgent;
import ips.media.event.MediaDurationAvailableEvent;
import ips.media.event.MediaEndEvent;
import ips.media.event.MediaErrorEvent;
import ips.media.event.MediaOpenEvent;
import ips.media.event.MediaStartedEvent;
import ips.media.event.MediaStoppedEvent;
import ips.media.io.StreamSource;
import ips.media.video.NativeWindowHandle;
import ips.media.video.VideoFormat;
import ips.media.video.VideoTrackListener;

public class GstMediaDecodingPlayer implements MediaDecodingPlayer,SeekControl,FrameStepControl, Runnable {

	private static final boolean DEBUG=false;
	
	private Thread gMainLoop;
	private volatile boolean loopRunning=false;
	private volatile boolean open=false;
	private volatile boolean running=false;
	
	private ByteBuffer nativePointer;

//	private File mediaFile;
	private Dimension videoSize;
	private VideoFormat videoFormat;

	public native ByteBuffer init();
	
	public native void configureVideoSink(ByteBuffer np,long x11Drawable);
	public native void setVideoSinkWindow(ByteBuffer np,long x11Drawable);
	public native void setAudioSink(ByteBuffer np,ByteBuffer audioSinkNativePointer);
	public native void gMainLoopRun(ByteBuffer np);
	public native void open(ByteBuffer np, String mediaFilePath);

	public native boolean start(ByteBuffer np);

	public native boolean stop(ByteBuffer np);
	public native boolean seek(ByteBuffer np,long reqPos);
	public native void frameStep(ByteBuffer np,long frames);
	public native boolean rewind(ByteBuffer np);
	public native long getDuration(ByteBuffer np);
	public native long getPosition(ByteBuffer np);
	public native void close(ByteBuffer np);
	public native void clear(ByteBuffer np);
	public native void release(ByteBuffer np);

	private MediaControlEventTransferAgent mceta;
	// private StreamSource mediaStreamSource;
	private File nativeMediaFile;
	
	private volatile byte[] imgDataBuffer=null;
	
	private Mixer.Info audioDeviceInfo;
	private SourceDataLine sourceDataLine;
	private SourceDataLine sdl;
	
	private long audioFramePositionSync=0;
	private volatile long audioBytesWritten;
	private boolean videosinkSet=false; 
	
	private boolean useJavaSound=false;



	private Vector<MediaControlListener> ctrlListeners = new Vector<MediaControlListener>();
	private Vector<VideoTrackListener> listeners = new Vector<VideoTrackListener>();

	private volatile boolean stopReq;
	private ComponentColorModel ccm;
	private Hashtable<?, ?> ht;
	private PixelInterleavedSampleModel csm;
	private BufferedImage bi;
    private WritableRaster raster;
	private DataBuffer dataBuf;
	private Component awtVideoComponent;
	private volatile long audioFramePosition; 
	private volatile long audioTimeSync=-1;
	
	private final Runnable closeRunnable=new Runnable() {
		public void run() {
			close();
		}
	};

    private boolean audioSinkSet=false;
    
    private NativeWindowHandle nativeWindowHandle=null;

	public GstMediaDecodingPlayer() {
	    super();
	    mceta = new MediaControlEventTransferAgent();

	    nativePointer = init();
	    ccm = new ComponentColorModel(

	            ColorSpace.getInstance(ColorSpace.CS_sRGB),
	            // ColorSpace.getInstance(ColorSpace.CS_LINEAR_RGB),
	            true, false, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
	    ht = new Hashtable();
	   
	}

	public byte[] setVideoSize(int width, int height,int parWidth,int parHeight) {
		videoSize = new Dimension(width, height);
		videoFormat = new VideoFormat(width, height);
		if(parWidth!=parHeight){
			// pixel size not quadratic
			videoFormat.setPixelAspectRatio(new Dimension(parWidth,parHeight));
		}
		csm = new PixelInterleavedSampleModel(
				DataBuffer.TYPE_BYTE, videoSize.width, videoSize.height, 4,
				videoSize.width * 4, new int[] { 0,1,2,3 });
		
		int bufSize=videoSize.width* videoSize.height * 4;
		imgDataBuffer=new byte[bufSize];
		dataBuf = new DataBufferByte(imgDataBuffer, bufSize);
        
        raster = Raster.createWritableRaster(csm, dataBuf,
                new Point(0, 0));
//        bi = new BufferedImage(width,height,BufferedImage.TYPE_4BYTE_ABGR);
        bi=new BufferedImage(ccm, raster, false, null);
		//bi.setAccelerationPriority(1.0f);
		for (VideoTrackListener vtl : listeners) {
			vtl.setVideoFormat(videoFormat);
		}
		return imgDataBuffer;
	}


	public void render() {
	  
		////	public void render(ByteBuffer data){
		//      if (data != null) {
		////              byte[] dataArr=new byte[data.capacity()];
		////              data.get(dataArr);
		//
		//			DataBuffer dataBuf = new DataBufferByte(data, videoSize.width
		//					* videoSize.height * 4);
		//			
		//			WritableRaster wr = Raster.createWritableRaster(csm, dataBuf,
		//					new Point(0, 0));
		//			bi.setData(wr);

//		bi.setData(raster);
		for (VideoTrackListener vtl : listeners) {
			//vtl.decodedVideoFrame(bi.getData().getDataBuffer().);
		}
		//			bi.flush();
		//		}

	}

//	public boolean audioOpen(int sampleRate,int bytesPerSample,int channels,boolean bigEndian,int bufSize){
//	    if(DEBUG){
//	        System.out.println("Audio open");
//	    }
//		int sampleSizeInBits=bytesPerSample*8/channels;
//		AudioFormat audioFormat=new AudioFormat(sampleRate, sampleSizeInBits, channels, true, bigEndian);
//		audioBytesWritten=0;
//		audioFramePosition=0;
////		audioFramePositionSync=
//		if(audioDeviceInfo==null){
//		    try {
//                sourceDataLine=AudioSystem.getSourceDataLine(audioFormat);
//            } catch (LineUnavailableException e) {
//               return false;
//            }
//		}else{
//		    try {
//                AudioSystem.getSourceDataLine(audioFormat, audioDeviceInfo);
//            } catch (LineUnavailableException e) {
//               return false;
//            }
//		}
//		if(sourceDataLine==null){
//		    return false;
//		}
//		
//			try {
//			   
//				if(DEBUG)System.out.println("Audio "+audioFormat+" Pref Buffer size "+bufSize);
//				sdl.open(audioFormat,bufSize);
//				int bufferSize=sdl.getBufferSize();
//				if(DEBUG)System.out.println("Buffer size "+bufferSize+" "+bufferSize/(audioFormat.getFrameSize()*audioFormat.getSampleRate()));
//				 if(!sdl.isActive()){
//			            sdl.start();
//			            if(DEBUG)System.out.println("SDL started");
//			      }
//				 return true;
//			} catch (LineUnavailableException e) {
//				
//				return false;
//			}
//		
//	}
//	
//	public int audioWrite(byte[] buf){
////	    System.out.println("Audio write "+buf.length+" bytes.");
////		if(audioStart==null){
////			audioStart=System.currentTimeMillis()+100;
////		}
//		
////		while(audioBytesDelay()>8000){
////			try {
////				Thread.sleep(50);
////			} catch (InterruptedException e) {
////				// TODO Auto-generated catch block
////				e.printStackTrace();
////			}
////		}
////	    int written=0;
////		int avail=sdl.available();
////		int c=0;
////		while(avail<buf.length && c<40){
////		    try {
////                Thread.sleep(50);
////                c++;
////            } catch (InterruptedException e) {
////                // TODO Auto-generated catch block
////                e.printStackTrace();
////            }
////		    avail=sdl.available();
////		}
//		int written=sdl.write(buf, 0, buf.length);
////		written=buf.length;
//		audioBytesWritten+=written;
//		
////		if(!sdl.isActive() && sdl.getBufferSize()-sdl.available()>=512){
//		int filled=sdl.getBufferSize()-sdl.available();
////		 System.out.println("Audio written "+written+" bytes,filled: "+filled);
//		return written;
//	}
//	
//	public int audioDelay(){
////		if(audioStart==null)return 0;
////		long playTime=System.currentTimeMillis()-audioStart;
////	
////		long framesPlayed=playTime*48;
////		long bytesPlayed=framesPlayed*4;
//////		long bytesWritten =audioFramesWritten*4;
////		long delay= audioFramesWritten-bytesPlayed;
////		if(delay<0){
////			delay=0;
////		}
//////		System.out.println("Delay: "+delay+" "+playTime/1000);
////		return delay;
//		AudioFormat af=sdl.getFormat();
//	
//		int frameSize=af.getFrameSize();
////		float frameRate=af.getFrameRate();
////		int sSize=af.getFrameSize()/af.getChannels();
////		int bufSize=sdl.getBufferSize();
////		int avail=sdl.available();
////		int filled=bufSize-avail;
//		// samples (frames?) queued
//		int delay;
////		delay=filled/frameSize;
//		long framePos=sdl.getLongFramePosition();
//		long retFramePos=framePos-audioFramePositionSync;
////		if(framePos!=audioFramePosition){
//////		    System.out.println("delay: framepos changed: "+framePos);
////		    audioTimeSync=System.currentTimeMillis();
////		}else{
////		   
////		    if(audioTimeSync!=-1){
////		        long now=System.currentTimeMillis();
////		        long timeDiff=now-audioTimeSync;
////		        long addFrames=(long)((timeDiff*frameRate)/1000.0);
////		        retFramePos=framePos+addFrames;
//////		        System.out.println("delay: framepos add frames: "+addFrames);
////		    }else{
//////		        System.out.println("delay: framepos no change");
////		    }
////		}
////		audioFramePosition=framePos;
////		long delay=(bufSize-avail)/frameSize;
//		delay=(int) (audioBytesWritten/frameSize -retFramePos);
//
//		return delay;
//	}
//	
//	public void audioReset(){
//	    if(DEBUG)System.out.println("Audio reset...");
//		if(sdl!=null){
//		    sdl.stop();
//			sdl.flush();
//			 if(DEBUG)System.out.println("Audio flushed.");
//			 // sync
//			 audioFramePositionSync=sdl.getLongFramePosition();
//			 audioBytesWritten=0;
//			sdl.start();
//			if(DEBUG)System.out.println("Audio restarted.");
//		}
//	}
//	
//	public boolean audioUnprepare(){
//	    
//	    return  true;
//	}
//	
//	public boolean audioClose(){
//		if(DEBUG)System.out.println("Audio close...");
//		if(sdl!=null){
//			sdl.stop();
//			sdl.close();
//		}
//		if(DEBUG)System.out.println("Audio closed.");
//		return true;
//
//	}
	/**
	 * Add video track listener.
	 * 
	 * @param vtl
	 *            video track listener
	 */
	public void addVideoTrackListener(VideoTrackListener vtl) {

		synchronized (listeners) {
			if (vtl != null && !listeners.contains(vtl)) {
				listeners.addElement(vtl);
			}
		}
	}

	/**
	 * Remove video track listener.
	 * 
	 * @param vtl
	 *            video track listener
	 */
	public synchronized void removeVideoTrackListener(VideoTrackListener vtl) {
		if (vtl != null) {
			listeners.removeElement(vtl);
		}
	}

	protected void finalize() {
		release(nativePointer);

	}

	// public void renderTestJ() {
	// renderTest(nativePointer);
	// thread.start();
	// }
	public NativeMediaResult open() throws NativeMediaSystemException {
	    if(open) return new NativeMediaResult(NativeMediaResult.Type.SUCCESS);
	    if(nativeMediaFile==null){
	        throw new NativeMediaSystemException("No media file set!");
	    }
	    long x11Drawable=0;
	   
	    if(nativeWindowHandle!=null){
	        if(nativeWindowHandle instanceof X11WindowHandle){
	            X11WindowHandle x11windowHandle=(X11WindowHandle)nativeWindowHandle;
	            x11Drawable=x11windowHandle.getDrawable();
	        }else {
	            throw new NativeMediaSystemException("Native window handle not suitable for X11!");
	        }
	    }else if(awtVideoComponent!=null){
	        synchronized (awtVideoComponent.getTreeLock()){
	            x11Drawable=nativeWindowHandleX11Drawable(awtVideoComponent);
	        }
	    }
	    
	     if(!nativeMediaFile.exists()){
	         throw new NativeMediaSystemException("Media file "+nativeMediaFile+" not found!");
	    }
	    
	   
	    if(!videosinkSet){
	        configureVideoSink(nativePointer, x11Drawable);
	        videosinkSet=true;
	    }

	    if(x11Drawable!=0){
	        setVideoSinkWindow(nativePointer, x11Drawable);
	    }
	    
	    if(!audioSinkSet && useJavaSound){
	        AudioSink audioSink=new AudioSink();
	        setAudioSink(nativePointer,audioSink.nativePointer());
	        audioSinkSet=true;
	    }
		String mediaFileUriStr=null;
		if(nativeMediaFile!=null){
			URI mediFileUri=nativeMediaFile.toURI();
			mediaFileUriStr=mediFileUri.toString();
		}
		
		gMainLoop=new Thread(this,"Glib-main-loop");
		loopRunning=true;
		
		
		gMainLoop.start();
		open(nativePointer,mediaFileUriStr);
		open=true;
		return new NativeMediaResult(NativeMediaResult.Type.SUCCESS);
	}
	
	public void error(String errMsg){
		System.err.println(errMsg);
		mceta.fireEvent(new MediaErrorEvent(this,new NativeMediaSystemException(errMsg)));
		
		// multiple subsequent error messages may occur on the bus
		// so we cannot close the player here
	}
	
	public void opened(){
	    if(DEBUG)System.out.println("Opened");
		mceta.fireEvent(new MediaOpenEvent(this));
	}
	
	public void started(){
	    if(DEBUG)System.out.println("Started");
		mceta.fireEvent(new MediaStartedEvent(this));
	}
	

	public void start() {
	    running=true;
	    start(nativePointer);
	    if(sdl!=null){
	        sdl.start();
	    }
	    for (VideoTrackListener vtl : listeners) {
	        vtl.activate();
	    }
	}

	public void stop() {
		if (stop(nativePointer)) {
			mceta.fireEvent(new MediaStoppedEvent(this));
		}
		if(sdl!=null){
			sdl.stop();
			
		}
		running=false;
	}
	
	public void durationAvailable(long duration){
	    mceta.fireEvent(new MediaDurationAvailableEvent(this, duration));
	}

	public void close() {
	    if(!open)return;
		loopRunning=false;
		close(nativePointer);
		if(gMainLoop!=null){
		if(DEBUG)System.out.print("Join g_main_loop...");
		try {
			gMainLoop.join(10000);
		} catch (InterruptedException e) {
			// OK
		}
		if (DEBUG)System.out.println("Joined.");
		}
//		videosinkSet=false;
//		audioSinkSet=false;
		clear(nativePointer);
		open=false;
		mceta.fireEvent(new MediaCloseEvent(this));
	}

	public void run() {
		while (loopRunning) {
			gMainLoopRun(nativePointer);
		}
		if (DEBUG)System.out.println("Java Glib main loop end.");
	}

	public void pause() {
		if(running){
		    stop();
		}else{
		    start();
		}

	}
	
	public void endOfStream(){
		if(DEBUG)System.out.println("EOS!");
		mceta.fireEvent(new MediaEndEvent(this));
	}

	public void addMediaControlListener(
			MediaControlListener mediaControlListener) {
		mceta.addListener(mediaControlListener);

	}

	public void removeMediaControlListener(
			MediaControlListener mediaControlListener) {
		mceta.removeListener(mediaControlListener);
	}

	public void rewind() {
		rewind(nativePointer);
	}

	public void setNativeMediaFile(File mediaFile) {
		this.nativeMediaFile=mediaFile;
		
	}

	public void stepForward() {
		// TODO Auto-generated method stub
		
	}
	
    
    public void setAWTVideoComponent(Component videoCanvas) throws NativeMediaSystemException {
        if(videoCanvas.isDisplayable() && videoCanvas.isLightweight()){
            throw new NativeMediaSystemException("Lightweight (Swing) components not supported! Please use AWT Canvas.");
        }
        this.awtVideoComponent = videoCanvas;
    }
    public boolean isNativeVideoComponentRenderingSupported() {
        // Gstreamer Linux (UNIX?) X11 only
        return true;
    }

    public Mixer.Info getAudioDeviceInfo() {
        return audioDeviceInfo;
    }

    public void setAudioDeviceInfo(Mixer.Info audioDeviceInfo) {
        this.audioDeviceInfo = audioDeviceInfo;
    }

    public boolean isJavaSoundSupported() {
       
        return true;
    }

    public void setUseJavaSound(boolean useJavaSound) {
        this.useJavaSound=useJavaSound;
        
    }

    public boolean isMediaStreamSourceSupported() {
        return false;
    }
    
    public StreamSource getMediaStreamSource() {
        // not supported
        return null;
    }

    public void setMediaStreamSource(StreamSource mediaStreamSource) {
        // not supported
        nativeMediaFile=null;
    }

    @Override
    public void frameStep(long frames) {
        frameStep(nativePointer,frames);
    }

    @Override
    public boolean backwardSteppingSupported() {
        return false;
    }

	@Override
	public long getPositionNs() {
		return getPosition(nativePointer);
	}

    @Override
    public boolean seek(long reqPosNs) {
        return seek(nativePointer, reqPosNs);
    }

   

    @Override
    public void setNativeVideoCanvasSize(double width, double height) {
        // TODO Auto-generated method stub
        
    }

    @Override
    public void setSilent(boolean silent) {
        // TODO Auto-generated method stub
        
    }

    @Override
    public boolean isOpen() {
        return open;
    }

    @Override
    public boolean isRunning() {
        return running;
    }

    @Override
    public boolean requiresDisplayableToRun() {
        
        return true;
    }

    @Override
    public void setVideoComponentNativeWindowHandle(NativeWindowHandle nativeWindowHandle)
            throws NativeMediaSystemException {
        
        
    }
    
    private native long nativeWindowHandleX11Drawable(Component heavyWeightComponent);
   

}
